import abc
import copy
import json
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    Union,
)

from bson import json_util
from bson.objectid import ObjectId
from pydantic import BaseModel, ValidationError

from .collections import get_collection
from .datastructures import BaseCartConfig, ItemConfig
from .exceptions import CartException
from .product_items import ProductItems


class BaseCart(
    metaclass=abc.ABCMeta,
):
    config_cls: ClassVar[Type[BaseCartConfig]]
    items: ProductItems
    coupon_code: Optional[str]
    _cache_cart: MutableMapping[str, Any]
    _subtotal: int
    _fee: int
    _discount: int

    _tax_ratio = 1.05

    # 結帳時訂單折扣後價格使用點數的比例
    _use_points_ratio = 1
    # 點數對金額的比例
    _points_ratio: int = 1
    # 金額對點數的比例
    _gift_points_ratio: int = 1

    INITIAL_CART = {
        "items": [],
        "extra_items": [],
        "discount_items": [],
        "total": 0,
        "subtotal": 0,
        "sales": 0,
        "tax": 0,
        "fee": 0,
        "discount": 0,
    }

    @property
    def subtotal(self) -> int:
        return self._subtotal

    @property
    def fee(self) -> int:
        return self._fee

    @property
    def discount(self) -> int:
        return self._discount

    @property
    def total(self) -> int:
        return self._subtotal + self._fee - self._discount

    @property
    def sales(self) -> int:
        sales_amount = sum(item.sales_amount for item in self.items)
        sales_fee = round(self._fee / self._tax_ratio, 7)
        sales_discount = round(self._discount / self._tax_ratio, 7)
        return round(sales_amount + sales_fee - sales_discount)

    @property
    def tax(self) -> int:
        return self.total - self.sales

    def __getattr__(self, item: str) -> Any:
        if item in self.config_cls.__fields__:
            _model_field = self.config_cls.__fields__[item]
            return self._cache_cart.get(item, _model_field.default)
        return self.__dict__[item]

    def __setattr__(self, key: str, value: Any) -> None:
        if key in self.config_cls.__fields__:
            self._cache_cart[key] = value
        else:
            self.__dict__[key] = value

    def __init__(
        self,
        identity: str,
        coupon: Optional[str] = None,
        points: int = 0,
        gift_points: int = 0,
        use_full_points: bool = False,
        initial_raise: bool = True,
        initial_clear: bool = False,
        extra_cart_query: Dict[str, Any] = {},
    ):
        self.coupon_code = coupon
        self._cache_cart = self._load_cart(identity, extra_cart_query)
        self.points = points
        self.gift_points = gift_points
        self.use_full_points = use_full_points
        self._user = None
        self._user_profile = None

        if not identity.startswith("tmp_"):
            self._user = get_collection("user").find_one({"_id": ObjectId(identity)})
            self._user_profile = get_collection("user_profile").find_one(
                {"owner": identity}
            )

        try:
            self.items = self.product_multi_item_cls(
                products=[
                    {
                        "id": item["product"]["id"],
                        "config": item["config"],
                    }
                    for item in self._cache_cart["items"]
                ],
            )
        except Exception as exc:
            # we have to manually clear the cart in this case, otherwise the user won't be able to get_cart()
            if initial_clear:
                self.items = self.product_multi_item_cls(products=[])
                self.empty_cart()

            if initial_raise:
                raise CartException(exc)

        self.calc()

    def get_cart_ins(self) -> MutableMapping[str, Any]:
        cart_dict = copy.deepcopy(self._cache_cart)
        return cart_dict

    def to_dict(self) -> MutableMapping[str, Any]:
        cart_dict = copy.deepcopy(self._cache_cart)
        cart_dict = json.loads(json_util.dumps(cart_dict))
        return cart_dict

    def add_item(self, product_id: str, config: ItemConfig) -> MutableMapping[str, Any]:
        self.items.add_item(product_id, config)
        self.calc()
        self.save_cart()
        return self.to_dict()

    def delete_item(self, index: int) -> MutableMapping[str, Any]:
        self.items.delete_item(index)
        self.calc()
        self.save_cart()
        return self.to_dict()

    def edit_item(self, index: int, config: ItemConfig) -> MutableMapping[str, Any]:
        self.items.edit_item(index, config)
        self.calc()
        self.save_cart()
        return self.to_dict()

    def edit_config(
        self, config: Union[Dict[str, Any], BaseModel]
    ) -> MutableMapping[str, Any]:
        if not isinstance(config, BaseModel):
            try:
                config = self.config_cls.validate_optional(config)
            except ValidationError:
                raise CartException("validation_error")

        for key, value in config.dict(exclude_unset=True).items():
            setattr(self, key, value)

        self.calc()
        self.save_cart()
        return self.to_dict()

    def empty_cart(self) -> MutableMapping[str, Any]:
        self.items.clean_items()
        self.coupon_code = None
        self._total = 0
        self._sales = 0
        self._fee = 0
        self._tax = 0
        self.points = 0
        self.gift_points = 0
        self.order_note = ""
        self.use_full_points = False
        self.calc()
        self.save_cart()
        return self.to_dict()

    def save_cart(self) -> None:
        get_collection("cart").find_one_and_update(
            {"_id": self._cache_cart["_id"]}, {"$set": self._cache_cart}
        )

    def calc(self) -> None:
        subtotal = self.items.total
        product_items = self.items.to_dict()
        fee, fee_items = self._calc_fee()
        discount, discount_items = self._calc_discounts(
            subtotal=subtotal, coupon_code=self.coupon_code
        )

        (
            addon_discount,
            addon_discount_items,
            addon_fee,
            addon_fee_items,
        ) = self._calc_addon(
            subtotal=subtotal,
            product_items=product_items,
            fee=fee,
            fee_items=fee_items,
            discount=discount,
            discount_items=discount_items,
        )

        self._subtotal = subtotal
        self._fee = fee + addon_fee
        self._discount = discount + addon_discount

        self._cache_cart["subtotal"] = self.subtotal
        self._cache_cart["fee"] = self.fee
        self._cache_cart["discount"] = self.discount
        self._cache_cart["total"] = self.total
        self._cache_cart["sales"] = self.sales
        self._cache_cart["tax"] = self.tax
        self._cache_cart["items"] = product_items
        self._cache_cart["extra_items"] = fee_items + addon_fee_items
        self._cache_cart["discount_items"] = discount_items + addon_discount_items

    def _load_cart(
        self,
        identity: str,
        extra_cart_query: Dict[str, Any] = {},
    ) -> MutableMapping[str, Any]:
        cart = get_collection("cart").find_one(
            {"owner": str(identity), **extra_cart_query}
        )
        if not cart:
            cart = {
                **extra_cart_query,
                **self.INITIAL_CART,
                **self.config_cls.validate_optional(self._init_cart_extra()).dict(),
                "owner": str(identity),
            }
            cart_result = get_collection("cart").insert_one(cart)
            cart["_id"] = cart_result.inserted_id
        return cart  # type: ignore

    @abc.abstractmethod
    def _init_cart_extra(self) -> Dict[str, Any]:
        raise NotImplementedError("_init_cart_extra")

    @abc.abstractmethod
    def _calc_fee(self) -> Tuple[int, List[Dict[str, Any]]]:
        raise NotImplementedError("_calc_fee")

    @abc.abstractmethod
    def _calc_discounts(
        self, subtotal: int, coupon_code: Optional[str] = None
    ) -> Tuple[int, List[Dict[str, Any]]]:
        raise NotImplementedError("_calc_discounts")



from typing import Any, Dict, List

from fastel.config import SdkConfig
from fastel.utils import requests


def validation_request(
    issuer: str, identity: str, ext_data: Dict[str, Any] = {}
) -> Dict[str, Any]:
    url = f"{SdkConfig.auth_host}/validation/request?client_id={SdkConfig.client_id}&client_secret={SdkConfig.client_secret}&issuer={issuer}"
    result = requests.post(
        url,
        json={"identity": identity, "ext_data": ext_data},
    ).json()
    assert isinstance(result, dict)
    return result


def validation_confirm(validation_id: str, code: str) -> Dict[str, Any]:
    url = f"{SdkConfig.auth_host}/validation/confirm/{validation_id}/{code}?client_id={SdkConfig.client_id}"
    result = requests.post(url, json={}).json()
    assert isinstance(result, dict)
    return result


def gen_token_request(
    identity: str, groups: List[str] = [], extra: Dict[str, Any] = {}
) -> Dict[str, Any]:
    url = f"{SdkConfig.auth_host}/jwt/server/encode?client_id={SdkConfig.client_id}&client_secret={SdkConfig.client_secret}"
    result = requests.post(
        url,
        json={"id": identity, "groups": groups, "ext": extra},
    ).json()
    assert isinstance(result, dict)
    return result


def revoke_token_request(identity: str) -> Dict[str, Any]:
    url = f"{SdkConfig.auth_host}/jwt/revoke?client_id={SdkConfig.client_id}&client_secret={SdkConfig.client_secret}"
    result = requests.post(url, json={"sub": identity}).json()
    assert isinstance(result, dict)
    return result



from enum import Enum
from typing import Any, Dict

from pydantic import BaseModel


class InvoiceStatus(str, Enum):
    success = "success"
    failure = "failure"
    pending = "pending"


class InvoiceIssueResp(BaseModel):
    status: InvoiceStatus
    invoice_number: str
    invoice_message: str
    invoice_detail: Dict[str, Any]
    invoice_random_no: str



import time
from typing import Literal, Optional

from pydantic import BaseModel, EmailStr, Field


def timestamp_generate() -> str:
    return str(int(time.time()))


class CheckoutModel(BaseModel):
    MerchantOrderNo: str
    Amt: int
    ItemDesc: str = "Online Payment To Neweb pay"
    Email: EmailStr
    CREDIT: Optional[Literal[0, 1]]
    ANDROIDPAY: Optional[Literal[0, 1]]
    SAMSUNGPAY: Optional[Literal[0, 1]]
    LINEPAY: Optional[Literal[0, 1]]
    ImageUrl: Optional[str]
    CreditRed: Optional[Literal[0, 1]]
    CREDITAE: Optional[Literal[0, 1]]
    UNIONPAY: Optional[Literal[0, 1]]
    WEBATM: Optional[Literal[0, 1]]
    VACC: Optional[Literal[0, 1]]
    CVS: Optional[Literal[0, 1]]
    BARCODE: Optional[Literal[0, 1]]
    ALIPAY: Optional[Literal[0, 1]]
    P2G: Optional[Literal[0, 1]]
    CVSCOM: Literal[0, 1, 2, 3] = 0

    NotifyURL: Optional[str]
    ReturnURL: Optional[str]
    CustomerURL: Optional[str]

    MerchantID: Optional[str]
    Version: Optional[str] = "2.0"
    RespondType: Optional[str] = "JSON"
    TimeStamp: str = Field(default_factory=timestamp_generate)

    LgsType: Literal["B2C", "B2B"] = "B2C"



from typing import Any, Dict, List, Tuple, Union

import boto3
from bs4 import BeautifulSoup
from fastapi import Response
from fastapi.requests import Request
from fastapi.responses import HTMLResponse
from requests import Response as ResponseType

from fastel.cart.datastructures import OverseaSubtype
from fastel.collections import get_collection, get_site_config
from fastel.config import SdkConfig
from fastel.exceptions import APIException
from fastel.logistics.common.models import (
    LogisticsDataModel,
    LogisticsStatus,
    LogisticsTrackStatusModel,
    LogisticTypes,
)
from fastel.logistics.ecpay.gateway import EcpayLogistics
from fastel.logistics.ecpay.models import (
    CvsMapSubTypeOptions,
    CVSSubtypes,
    EcpayCVSMapModel,
    EcpayLogisticsCallbackModel,
    EcpayLogisticsModel,
    EcpayPrintOrderInfo,
)
from fastel.logistics.sf.gateway import SFLogistics
from fastel.logistics.sf.models import SFLogisticsModel
from fastel.payment.common.models.order import Order, ProductItem
from fastel.utils import requests, timestamp_postfix


class EcpayLogisticsCreate:
    def __init__(self, callback_url: str) -> None:
        self.gateway = EcpayLogistics(
            merchant_id=SdkConfig.ecpay_logistics_merchant_id,
            hash_key=SdkConfig.ecpay_logistics_hash_key,
            hash_iv=SdkConfig.ecpay_logistics_hash_iv,
            stage=SdkConfig.stage,
        )
        self.server_reply_url = callback_url

    def __call__(
        self, order: Dict[str, Any], extra_data: Dict[str, Any] = {}
    ) -> LogisticsDataModel:
        validated = self.validate(order)
        parser = self.parse_logistics(validated, extra_data)
        gateway_resp = self.gateway.create_logistics(data=parser)
        result = self.parse_gateway_resp(gateway_resp=gateway_resp)
        return result

    @staticmethod
    def validate(order: Dict[str, Any]) -> Order:
        return Order.validate(order)

    @staticmethod
    def _replace_limit_name(name: str, limit: int) -> str:
        if len(name) > limit:
            return name[: limit - 3] + "..."
        return name

    def parse_logistics(
        self, order: Order, extra_data: Dict[str, Any]
    ) -> EcpayLogisticsModel:
        if order.total > 20000:
            APIException(
                status_code=400, error="validation_error", detail="物流訂單商品不能超過20000元新台幣"
            )
        counter = get_collection("counter").find_one_and_update(
            {"collection": "logistics", "order_id": order.order_id},
            {
                "$inc": {"idx": 1},
                "$setOnInsert": {"collection": "logistics", "order_id": order.order_id},
            },
            upsert=True,
        )
        if not counter:
            counter = {"idx": 0}
        common_props = dict(
            MerchantTradeNo=f"{order.order_id}_{counter['idx'] + 1}",
            GoodsName=f"訂單{order.order_id}",
            GoodsAmount=order.total,
            ReceiverName=self._replace_limit_name(name=order.receiver_name, limit=10),
            ReceiverEmail=order.receiver_email,
            ReceiverCellPhone=order.receiver_phone,
            ReceiverZipCode=order.receiver_zip,
            ReceiverAddress=(
                order.receiver_city + order.receiver_district + order.receiver_address
            ),
            ServerReplyURL=self.server_reply_url,
        )

        default_sender_info = dict(
            SenderName="忻旅科技",
            SenderPhone="02-77295130",
            SenderZipCode="10361",
            SenderCellPhone="0900000000",
            SenderAddress="台北市大同區民權西路136號10樓之5",
        )

        sender_info = get_site_config(
            key="logistics_sender_info", default=default_sender_info
        )
        sender_info["SenderName"] = self._replace_limit_name(
            name=sender_info["SenderName"], limit=10
        )
        data: EcpayLogisticsModel
        if order.logistics_type == LogisticTypes.HOME.value:
            data = EcpayLogisticsModel(
                **common_props,
                **sender_info,
                **extra_data,
                Distance=("00" if order.receiver_city == "台北市" else "01"),
                Temperature="0001",
            )
        elif order.logistics_type == LogisticTypes.CVS.value:
            data = EcpayLogisticsModel(
                **common_props,
                **sender_info,
                **extra_data,
                LogisticsType="CVS",
                LogisticsSubType=order.logistics_subtype,
                IsCollection="Y" if order.payment_subtype == "cod" else "N",
                CollectionAmount=order.total if order.payment_subtype == "cod" else 0,
                ReceiverStoreID=order.logistics_cvs_store_id,
            )
        elif order.logistics_type == LogisticTypes.OVERSEA.value:
            raise APIException(
                status_code=400,
                error="logistics_type_invalid",
                detail="ecpay logistics type OVERSEA is invalid",
            )
        else:
            raise APIException(
                status_code=400,
                error="validation_error",
                detail=f"not found logistics_type {order.logistics_type}",
            )
        return data

    @staticmethod
    def parse_gateway_resp(gateway_resp: Dict[str, Any]) -> LogisticsDataModel:
        if gateway_resp.get("error", {}):
            error = "error"
            error_msg = gateway_resp["error"]
            return LogisticsDataModel(
                provider="ecpay",
                order_id=error,
                logistics_id=error,
                logistics_type=LogisticTypes.UNKNOWN,
                logistics_subtype="",
                logistics_status=LogisticsStatus.error,
                logistics_message=error_msg,
                logistics_detail=gateway_resp,
            )

        return LogisticsDataModel(
            provider="ecpay",
            order_id=gateway_resp["MerchantTradeNo"].split("_")[0],
            logistics_id=gateway_resp["AllPayLogisticsID"],
            logistics_type=gateway_resp["LogisticsType"],
            logistics_subtype=gateway_resp["LogisticsSubType"],
            logistics_status=LogisticsStatus.pending,
            logistics_message=gateway_resp["RtnMsg"],
            logistics_detail=gateway_resp,
        )


class EcpayLogisticsCallback:
    def __init__(self) -> None:
        self.gateway = EcpayLogistics(
            merchant_id=SdkConfig.ecpay_logistics_merchant_id,
            hash_key=SdkConfig.ecpay_logistics_hash_key,
            hash_iv=SdkConfig.ecpay_logistics_hash_iv,
            stage=SdkConfig.stage,
        )
        self.response = "0|NOPE"

    def __call__(self, data: Dict[str, Any]) -> LogisticsDataModel:
        validated = self.validate(data)
        self.check_mac_data(data=validated)
        parsed = self.parse_payload(validated)
        return parsed

    async def parse_request(self, request: Request) -> Dict[str, Any]:
        data = await request.form()
        return requests.parse_formdata(data)

    def validate(self, data: Dict[str, Any]) -> EcpayLogisticsCallbackModel:
        return EcpayLogisticsCallbackModel.validate(data)

    def check_mac_data(self, data: EcpayLogisticsCallbackModel) -> None:
        mac_value = data.CheckMacValue
        result_data = data.dict(exclude_none=True)
        result_data.pop("CheckMacValue", None)

        result = self.gateway.cryptor.encrypt(result_data)
        if result == mac_value:
            self.response = "1|OK"

    def parse_payload(self, result: EcpayLogisticsCallbackModel) -> LogisticsDataModel:
        if result.LogisticsType == LogisticTypes.HOME.value:
            home_status_table = {
                "300": LogisticsStatus.pending,
                "310": LogisticsStatus.pending,
                "3001": LogisticsStatus.center_delivered,
                "3003": LogisticsStatus.delivered,
                "3006": LogisticsStatus.in_delivery,
            }
            try:
                print(result.RtnCode)
                logistics_status = home_status_table[result.RtnCode]
            except KeyError:
                logistics_status = LogisticsStatus.exception
        else:
            cvs_status_table: Dict[str, Any] = {
                CVSSubtypes.FAMIC2C: {
                    "300": LogisticsStatus.pending,
                    "310": LogisticsStatus.pending,
                    "3024": LogisticsStatus.center_delivered,
                    "3018": LogisticsStatus.store_delivered,
                    "3022": LogisticsStatus.delivered,
                },
                CVSSubtypes.UNIMARTC2C: {
                    "300": LogisticsStatus.pending,
                    "310": LogisticsStatus.pending,
                    "2030": LogisticsStatus.center_delivered,
                    "2073": LogisticsStatus.store_delivered,
                    "2067": LogisticsStatus.delivered,
                },
                CVSSubtypes.HILIFEC2C: {
                    "300": LogisticsStatus.pending,
                    "310": LogisticsStatus.pending,
                    "2001": LogisticsStatus.pending,
                    "2030": LogisticsStatus.center_delivered,
                    "3024": LogisticsStatus.center_delivered,
                    "2063": LogisticsStatus.store_delivered,
                    "3018": LogisticsStatus.store_delivered,
                    "2067": LogisticsStatus.delivered,
                    "3022": LogisticsStatus.delivered,
                },
                CVSSubtypes.OKMARTC2C: {
                    "300": LogisticsStatus.pending,
                    "310": LogisticsStatus.pending,
                    "2030": LogisticsStatus.center_delivered,
                    "2073": LogisticsStatus.store_delivered,
                    "3022": LogisticsStatus.delivered,
                },
                CVSSubtypes.FAMI: {
                    "300": LogisticsStatus.pending,
                    "310": LogisticsStatus.pending,
                    "3024": LogisticsStatus.center_delivered,
                    "3018": LogisticsStatus.store_delivered,
                    "3022": LogisticsStatus.delivered,
                },
                CVSSubtypes.UNIMART: {
                    "300": LogisticsStatus.pending,
                    "310": LogisticsStatus.pending,
                    "2030": LogisticsStatus.center_delivered,
                    "2073": LogisticsStatus.store_delivered,
                    "2067": LogisticsStatus.delivered,
                },
                CVSSubtypes.HILIFE: {
                    "300": LogisticsStatus.pending,
                    "310": LogisticsStatus.pending,
                    "2001": LogisticsStatus.pending,
                    "2030": LogisticsStatus.center_delivered,
                    "3024": LogisticsStatus.center_delivered,
                    "2063": LogisticsStatus.store_delivered,
                    "3018": LogisticsStatus.store_delivered,
                    "2067": LogisticsStatus.delivered,
                    "3022": LogisticsStatus.delivered,
                },
            }

            try:
                logistics_status = cvs_status_table[result.LogisticsSubType][
                    result.RtnCode
                ]
            except KeyError:
                logistics_status = LogisticsStatus.exception

        return LogisticsDataModel(
            provider="ecpay",
            order_id=result.MerchantTradeNo.split("_")[0],
            logistics_id=result.AllPayLogisticsID,
            logistics_type=result.LogisticsType,
            logistics_subtype=result.LogisticsSubType,
            logistics_status=logistics_status,
            logistics_message=result.RtnMsg,
            logistics_payment_no=result.CVSPaymentNo,
            logistics_validate_no=result.CVSValidationNo,
        )

    def respond(self, **kwargs: Any) -> Response:
        return Response(content=self.response)


class EcpayGetShippingNote:
    def __init__(self) -> None:
        self.gateway = EcpayLogistics(
            merchant_id=SdkConfig.ecpay_logistics_merchant_id,
            hash_key=SdkConfig.ecpay_logistics_hash_key,
            hash_iv=SdkConfig.ecpay_logistics_hash_iv,
            stage=SdkConfig.stage,
        )

    def __call__(self, data: Dict[str, Any]) -> Tuple[bytes, str]:
        validated = self.validate(data)
        parsed = self.parse_payload(validated)
        html, logistics_subtype = self.gateway.get_shipping_note(parsed)
        print(html)
        if logistics_subtype == "UniMartC2C":
            return html.encode(), "text/html"
        else:
            img_url = self.fetch_html_img(html)
            print("shipping note url:", img_url)
            img_url = img_url.replace("amp;", "")
            img_resp: ResponseType = requests.get(img_url)
            img_resp.raise_for_status()
            img = img_resp.content
            content_type = img_resp.headers.get("Content-Type", "image/png")
        return img, content_type

    @staticmethod
    def fetch_html_img(html: str) -> str:
        soup = BeautifulSoup(html)
        img_tag = soup.find("img")
        if not img_tag:
            form = soup.find("form")
            if not form:
                return ""
            else:
                return str(form.get("action", ""))
        else:
            return str(img_tag.get("src", ""))

    def validate(self, data: Dict[str, Any]) -> LogisticsDataModel:
        return LogisticsDataModel.validate(data)

    def parse_payload(self, result: LogisticsDataModel) -> EcpayPrintOrderInfo:
        if result.logistics_subtype == "UNIMARTC2C":
            payload = {
                "LogisticsSubType": "UniMartC2C",
                "AllPayLogisticsID": result.logistics_id,
                "CVSPaymentNo": result.logistics_payment_no,
                "CVSValidationNo": result.logistics_validate_no,
            }
        elif result.logistics_subtype in ["FAMIC2C", "OKMARTC2C", "HILIFEC2C"]:
            payload = {
                "LogisticsSubType": result.logistics_subtype,
                "AllPayLogisticsID": result.logistics_id,
                "CVSPaymentNo": result.logistics_payment_no,
            }
        elif result.logistics_subtype in [
            "FAMI",
            "UNIMART",
            "UNIMARTFREEZE",
            "HILIFE",
            "TCAT",
            "ECAN",
        ]:
            payload = {
                "LogisticsSubType": result.logistics_subtype,
                "AllPayLogisticsID": result.logistics_id,
            }

        return EcpayPrintOrderInfo.validate(payload)


class SFLogisticsCreate:
    def __init__(self) -> None:
        self.gateway = SFLogistics(
            merchant_id=SdkConfig.sf_merchant_id,
            aes_key=SdkConfig.sf_aes_key,
            app_key=SdkConfig.sf_app_key,
            secret=SdkConfig.sf_secret,
            stage=SdkConfig.stage,
        )

    def __call__(
        self, order: Dict[str, Any], extra_data: Dict[str, Any] = {}
    ) -> LogisticsDataModel:
        validated = self.validate(order)
        parsed = self.parse_logistics(validated, extra_data)
        gateway_resp = self.gateway.create_logistics(parsed)
        result = self.parse_gateway_resp(gateway_resp=gateway_resp, order=validated)
        return result

    @staticmethod
    def validate(order: Dict[str, Any]) -> Order:
        return Order.validate(order)

    @staticmethod
    def generate_items(items: List[ProductItem]) -> List[Dict[str, Any]]:
        result = []
        for item in items:
            data = {
                "name": item.name,
                "unit": "個",
                "amount": float(item.amount),
                "currency": "NTD",
                "quantity": float(item.config.qty),
                "originCountry": "TW",
            }
            result.append(data)
        return result

    def parse_logistics(
        self, order: Order, extra_data: Dict[str, Any]
    ) -> SFLogisticsModel:
        default_sender_info = dict(
            SenderName="忻旅科技",
            SenderPhone="02-77295130",
            SenderZipCode="10361",
            SenderCellPhone="0900000000",
            SenderAddress="台北市大同區民權西路136號10樓之5",
        )
        sender_info: Dict[str, Any] = get_site_config(
            key="logistics_sender_info", default=default_sender_info
        )
        counter = get_collection("counter").find_one_and_update(
            {"collection": "logistics", "order_id": order.order_id},
            {
                "$inc": {"idx": 1},
                "$setOnInsert": {"collection": "logistics", "order_id": order.order_id},
            },
            upsert=True,
        )
        if not counter:
            counter = {"idx": 0}
        if order.logistics_type == LogisticTypes.OVERSEA.value:
            sender_cert_no = sender_info.get("SenderCertNo", "")
            sender_cert_type = sender_info.get("SenderCertType", "001")
            receiver_country = OverseaSubtype[order.logistics_subtype].value
            receiver_post_code = extra_data.get("receiver_post_code", "")
            receiver_region_first = extra_data.get("receiver_region_first", "")
            receiver_region_second = extra_data.get("receiver_region_second", "")
            receiver_region_third = extra_data.get("receiver_region_third", "")
            receiver_address = extra_data.get("receiver_address", "")
            receiver_name = extra_data.get("receiver_name", "")
            receiver_email = extra_data.get("receiver_email", "")
            receiver_phone = extra_data.get("receiver_phone", "")
        else:
            sender_cert_no = ""
            sender_cert_type = ""
            receiver_country = "TW"
            receiver_post_code = order.receiver_zip
            receiver_region_first = "台灣省"
            receiver_region_second = order.receiver_city
            receiver_region_third = order.receiver_district
            receiver_address = order.receiver_address
            receiver_name = order.receiver_name
            receiver_email = order.receiver_email
            receiver_phone = order.receiver_phone
        payload = {
            "customerCode": SdkConfig.sf_merchant_id,
            "orderOperateType": 1,
            "customerOrderNo": f"{order.order_id}_{counter['idx']}",
            "interProductCode": extra_data.get("platform_code", "INT0005"),
            # 包裹數 *
            "parcelQuantity": extra_data.get("parcel_quantity", 1),
            # 聲明價值 * order
            "declaredValue": order.total,
            # 包裹总计声明价值币种
            "declaredCurrency": "NTD",
            # 寄件方式 0: 服務點自寄或自行聯繫快遞員 1: 上門收件
            "pickupType": "1",
            # 上門區間預約時間 yyyy-MM-dd HH:mm 如果pickupType 為 1 則必填
            "pickupAppointTime": extra_data.get("pickup_time"),
            # 收件時區
            "pickupAppointTimeZone": "Asia/Taipei",
            # 運單備註 *
            "remark": extra_data.get("note", ""),
            # 付款方式
            "paymentInfo": {
                # 付款方式 1 寄方付， 2 收方付， 3 第三方付
                "payMethod": "3",
                "taxPayMethod": "3",
                "payMonthCard": SdkConfig.sf_card_no,
                "taxPayMonthCard": SdkConfig.sf_card_no,
            },
            # 寄件人訊息
            "senderInfo": {
                # 寄件人名字
                "contact": sender_info.get("SenderName", ""),
                # 寄件國家/地區
                "country": "TW",
                # 郵編
                "postCode": sender_info.get("SenderZipCode", ""),
                # 州/省
                "regionFirst": "台灣省",
                # 城市
                "regionSecond": sender_info.get("SenderCity", "臺北市"),
                # 區
                "regionThird": sender_info.get("SenderDistrict", "大同區"),
                "address": sender_info.get("SenderAddress", ""),
                "email": sender_info.get("SenderEmail", ""),
                "cargoType": extra_data.get("cargo_type", 1),
                "telNo": sender_info.get("SenderPhone", ""),
                "certType": sender_cert_type,
                "certCardNo": sender_cert_no,
            },
            # 收件人訊息
            "receiverInfo": {
                # 寄件人名字
                "contact": receiver_name,
                # 寄件國家/地區
                "country": receiver_country,
                # 郵編
                "postCode": receiver_post_code,
                # 州/省
                "regionFirst": receiver_region_first,
                # 城市
                "regionSecond": receiver_region_second,
                # 區
                "regionThird": receiver_region_third,
                "address": receiver_address,
                "email": receiver_email,
                "cargoType": extra_data.get("cargo_type", 1),
                "phoneNo": receiver_phone,
            },
            # 包裹訊息
            "parcelInfoList": self.generate_items(order.items),
        }

        return SFLogisticsModel.validate(payload)

    @staticmethod
    def parse_gateway_resp(
        gateway_resp: Dict[str, Any], order: Order
    ) -> LogisticsDataModel:
        resp_data = gateway_resp.get("data", {})
        if gateway_resp.get("success", False):
            logistics_id = resp_data.get("sfWaybillNo", "")
            logistics_status = LogisticsStatus.pending
            logistics_message = "已收到訂單資料"

        else:
            logistics_id = ""
            logistics_status = LogisticsStatus.error
            logistics_message = gateway_resp.get("msg", "")

        return LogisticsDataModel(
            provider="sf",
            order_id=resp_data.get("customerOrderNo", ""),
            logistics_id=logistics_id,
            logistics_type=order.logistics_type,
            logistics_subtype=order.logistics_subtype,
            logistics_status=logistics_status,
            logistics_message=logistics_message,
        )


class SFLogisticsStatus:
    def __init__(self) -> None:
        self.gateway = SFLogistics(
            merchant_id=SdkConfig.sf_merchant_id,
            aes_key=SdkConfig.sf_aes_key,
            app_key=SdkConfig.sf_app_key,
            secret=SdkConfig.sf_secret,
            stage=SdkConfig.stage,
        )

    def __call__(self, logistics_id: str) -> LogisticsTrackStatusModel:
        parsed = self.parse_data(logistics_id)
        gateway_resp = self.gateway.query_track(parsed)
        result = self.parse_gateway_resp(gateway_resp=gateway_resp)
        return result

    @staticmethod
    def parse_data(logistics_id: str) -> Dict[str, Any]:
        data = {"customerCode": SdkConfig.sf_merchant_id, "sfWaybillNo": logistics_id}
        return data

    @staticmethod
    def parse_gateway_resp(gateway_resp: Dict[str, Any]) -> LogisticsTrackStatusModel:
        resp_data = gateway_resp.get("data", [])
        if resp_data:
            tracking_data = resp_data[-1]
        else:
            tracking_data = {}
        if gateway_resp.get("success", False):
            track_details: List[Dict[str, Any]] = tracking_data.get(
                "trackDetailItems", []
            )
            logistics_id = tracking_data.get("sfWaybillNo", "")
            logistics_message = ""
            if track_details:
                for detail in track_details:
                    logistics_message += detail.get("trackOutRemark", "") + ";"
            else:
                logistics_message = tracking_data.get("trackSummary", "")

        else:
            logistics_id = tracking_data.get("sfWaybillNo", "")
            logistics_message = gateway_resp.get("msg", "")

        return LogisticsTrackStatusModel(
            provider="sf",
            tracking_type="string",
            logistics_id=logistics_id,
            logistics_message=logistics_message,
            logistics_detail=gateway_resp,
        )


class EcpayLogisticsMap:
    def __init__(self, callback_url: str) -> None:
        self.gateway = EcpayLogistics(
            merchant_id=SdkConfig.ecpay_logistics_merchant_id,
            hash_key=SdkConfig.ecpay_logistics_hash_key,
            hash_iv=SdkConfig.ecpay_logistics_hash_iv,
            stage=SdkConfig.stage,
        )
        self.server_reply_url = callback_url

    def __call__(
        self, logistics_subtype: CvsMapSubTypeOptions, is_collection: bool = False
    ) -> HTMLResponse:
        parser = self.parse_map(logistics_subtype, is_collection)
        return HTMLResponse(self.gateway.cvs_map(data=parser))

    def parse_map(
        self, logistics_subtype: CvsMapSubTypeOptions, is_collection: bool = False
    ) -> EcpayCVSMapModel:

        return EcpayCVSMapModel(
            LogisticsType="CVS",
            LogisticsSubType=logistics_subtype,
            IsCollection="Y" if is_collection else "N",
            ServerReplyURL=self.server_reply_url,
        )


def get_logistics_map_method(
    provider: str = "ecpay", prefix_path: str = "logistics/map", state: str = ""
) -> EcpayLogisticsMap:
    if provider == "ecpay":
        return EcpayLogisticsMap(
            callback_url=f"{SdkConfig.api_host}/{prefix_path}/callback?state={state}"
        )
    raise ValueError("unrecognize logistics provider")


def get_logistics_create_method(
    provider: str = "ecpay",
    prefix_path: str = "logistics",
) -> Union[EcpayLogisticsCreate, SFLogisticsCreate]:
    if provider == "ecpay":
        return EcpayLogisticsCreate(
            callback_url=f"{SdkConfig.api_host}/{prefix_path}/{provider}/callback",
        )
    elif provider == "sf":
        return SFLogisticsCreate()
    raise ValueError("unrecognize logistics provider")


def get_callback_method(provider: str = "ecpay") -> EcpayLogisticsCallback:
    if provider == "ecpay":
        return EcpayLogisticsCallback()
    raise ValueError("unrecognize payment provider")


def get_logistics_status_method(provider: str = "sf") -> SFLogisticsStatus:
    if provider == "sf":
        return SFLogisticsStatus()
    raise ValueError("unrecognize logistics provider")


def get_shipping_note_method(provider: str = "ecpay") -> EcpayGetShippingNote:
    if provider == "ecpay":
        return EcpayGetShippingNote()
    raise ValueError("unrecognize logistics provider")


stepfn = boto3.client("stepfunctions", "ap-northeast-1")


async def process_callback(provider: str, request: Request) -> Response:
    callback_method = get_callback_method(provider)
    data = await callback_method.parse_request(request)
    callback_result = callback_method(data)
    # to avoid step function name duplicate error
    stepfn_name = f"{callback_result.order_id}_{callback_result.logistics_status}"
    stepfn.start_execution(
        stateMachineArn=SdkConfig.logistics_stepfn_arn,
        name=timestamp_postfix(stepfn_name),
        input=callback_result.json(),
    )
    return callback_method.respond(order_id=callback_result.order_id)



from typing import Any, Dict, Optional

from pydantic import BaseModel


class CallbackModel(BaseModel):
    client_id: Optional[str]
    payment_subtype: str
    payment_status: str
    payment_transaction_id: Optional[str]
    payment_getnum_id: Optional[str]
    payment_transaction_detail: Optional[Dict[str, Any]]
    payment_redirect_detail: Optional[Dict[str, Any]]
    payment_transaction_time: Optional[str]

    total: int
    order_id: str
    pay_deadline: Optional[str]

    code_no: Optional[str]

    bank_code: Optional[str]
    bank_account: Optional[str]

    barcode_1: Optional[str]
    barcode_2: Optional[str]
    barcode_3: Optional[str]


class LogisticCallbackModel(BaseModel):
    logistic_status: str
    logistic_type: str
    logistic_subtype: str
    logistic_id: str
    detail: Dict[str, Any]



import urllib.parse
from datetime import datetime
from typing import Any, Dict, Tuple

from requests import Response

from fastel.logistics.ecpay.models import (
    EcpayCVSMapModel,
    EcpayLogisticsModel,
    EcpayPrintOrderInfo,
)
from fastel.payment.cryptors import MD5Cryptor
from fastel.utils import TW, requests


class EcpayLogistics:
    def __init__(
        self,
        merchant_id: str = "2000933",
        hash_key: str = "XBERn1YOvpM9nfZc",
        hash_iv: str = "h1ONHk4P4yqbl5LK",
        stage: str = "stg",
    ) -> None:
        self.merchant_id = merchant_id
        self.hash_key = hash_key
        self.hash_iv = hash_iv
        self.stage = stage
        self.cryptor = MD5Cryptor(hash_key=self.hash_key, hash_iv=self.hash_iv)

    @property
    def logistics_url(self) -> str:
        if self.stage in ["prod", "production"]:
            return "https://logistics.ecpay.com.tw"
        return "https://logistics-stage.ecpay.com.tw"

    @property
    def unimart_url(self) -> str:
        if self.stage in ["prod", "production"]:
            return "https://epayment.7-11.com.tw/C2C/C2CWeb"
        return "https://epaymenttest.7-11.com.tw/C2C/C2CWeb"

    def create_logistics(self, data: EcpayLogisticsModel) -> Dict[str, Any]:
        url = self.logistics_url + "/Express/Create"
        data_dict = data.dict(exclude_none=True)
        now = datetime.now(TW)
        gateway_payload = {
            **data_dict,
            "MerchantID": self.merchant_id,
            "MerchantTradeDate": now.strftime("%Y/%m/%d %H:%M:%S"),
        }
        mac_value = self.cryptor.encrypt(gateway_payload)
        gateway_payload["CheckMacValue"] = mac_value

        print("[Logistics Create Req]", gateway_payload)
        resp: Response = requests.post(url, data=gateway_payload)
        resp.raise_for_status()
        rtn_code, real_resp = resp.text.split("|")
        print("[Logistics Create Resp]", rtn_code, real_resp)

        if rtn_code != "1":
            return {"error": real_resp}
        return dict(urllib.parse.parse_qsl(real_resp))

    @staticmethod
    def build_input(data: Dict[str, Any]) -> str:
        input_row = "\n".join(
            map(
                lambda k: f'<input type="hidden" name="{k[0]}" value="{k[1]}"><br>',
                data.items(),
            )
        )
        return input_row

    def cvs_map(self, data: EcpayCVSMapModel) -> str:
        url = self.logistics_url + "/Express/map"
        data_dict = data.dict(exclude_none=True)
        data_dict["MerchantID"] = self.merchant_id
        row = self.build_input(data=data_dict)
        template = f"""
        <html>
          <body>
              <form name='ecpay' method='post' action='{url}'>
                  {{}}
                  <input type='hidden' value='Submit'>
              </form>
              </body>
              <script type="text/javascript">ecpay.submit();
          </script>
        </html>
        """
        result = template.format(row)
        return result

    def get_shipping_note(self, data: EcpayPrintOrderInfo) -> Tuple[str, str]:
        request_body = data.dict(exclude_none=True)
        logistics_subtype = request_body.pop("LogisticsSubType")
        if logistics_subtype in [
            "FAMI",
            "UNIMART",
            "UNIMARTFREEZE",
            "HILIFE",
            "TCAT",
            "ECAN",
        ]:
            url = f"{self.logistics_url}/helper/printTradeDocument"
        elif logistics_subtype == "UniMartC2C":
            url = f"{self.unimart_url}/PrintC2CPinCode.aspx"
            headers = {
                "referer": f"{self.unimart_url}/C2C.aspx",
                "Content-Type": "application/x-www-form-urlencoded",
            }
            resp: Response = requests.post(
                url,
                headers=headers,
                data={
                    "PinCodeNumber": str(data.CVSPaymentNo) + str(data.CVSValidationNo)
                },
            )
            resp.raise_for_status()
            html: str = resp.content.decode()
            html = html.replace(
                'src="BarCode.ashx?CodeValue=',
                f'src="{self.unimart_url}/BarCode.ashx?CodeValue=',
            )
            html = html.replace(
                'src="QRCode.ashx?CodeValue=',
                f'src="{self.unimart_url}/QRCode.ashx?CodeValue=',
            )
            return html, logistics_subtype
        else:
            url = f"{self.logistics_url}/Express/Print{logistics_subtype}OrderInfo"

        request_body["MerchantID"] = self.merchant_id
        mac_value = self.cryptor.encrypt(request_body)
        request_body["CheckMacValue"] = mac_value
        print(url)
        resp = requests.post(url, data=request_body)
        resp.raise_for_status()
        return resp.content.decode(), logistics_subtype



import json
import uuid
from typing import Any, Dict

import requests

from fastel.payment.cryptors import HMACCryptor
from fastel.payment.linepay.models.checkout import (
    LinePayCheckoutModel,
    LinePayConfirmModel,
    LinePayConfirmResp,
    LinePayRequestResp,
    LinePayResponse,
    LinePayTransactionDetail,
)


class LinePay:
    def __init__(
        self,
        channel_id: str,
        channel_secret: str,
        stage: str = "stg",
    ):
        self.channel_id = channel_id
        self.channel_secret = channel_secret
        self.stage = stage
        self.cryptor = HMACCryptor(self.channel_secret)

    @property
    def payment_host(self) -> str:
        if self.stage in ["prod", "PROD"]:
            return "https://api-pay.line.me"
        return "https://sandbox-api-pay.line.me"

    def build_headers(
        self,
        uri: str,
        data_str: str,
    ) -> Dict[str, Any]:
        nonce = str(uuid.uuid4())
        signature = self.cryptor.encrypt(nonce, uri, data_str)
        return {
            "Content-Type": "application/json",
            "X-LINE-ChannelId": self.channel_id,
            "X-LINE-Authorization-Nonce": nonce,
            "X-LINE-Authorization": signature,
        }

    def checkout_request(self, checkout: LinePayCheckoutModel) -> LinePayRequestResp:
        body = checkout.dict()
        path = "/v3/payments/request"
        url = f"{self.payment_host}{path}"
        payload = json.dumps(body)
        headers = self.build_headers(path, payload)
        try:
            resp = requests.post(
                url=url,
                json=body,
                headers=headers,
            )
            resp.raise_for_status()
            resp_json = resp.json()
            resp_model = LinePayRequestResp.validate(resp_json)
        except requests.HTTPError as exc:
            resp_json = {"error": exc.response.json()}
            resp_model = LinePayRequestResp.validate(resp_json)

        assert isinstance(resp_model, LinePayRequestResp)
        return resp_model

    def checkout_confirm(self, confirm: LinePayConfirmModel) -> LinePayConfirmResp:
        body = confirm.dict()
        transaction_id = body.pop("transactionId", None)
        path = f"/v3/payments/{transaction_id}/confirm"
        payload = json.dumps(body)
        url = f"{self.payment_host}{path}"
        headers = self.build_headers(path, payload)
        try:
            resp = requests.post(
                url=url,
                json=body,
                headers=headers,
            )
            resp.raise_for_status()
            resp_json = resp.json()
            resp_model = LinePayConfirmResp.validate(resp_json)
        except requests.HTTPError as exc:
            resp_json = {"error": exc.response.json()}
            resp_model = LinePayConfirmResp.validate(resp_json)

        assert isinstance(resp_model, LinePayConfirmResp)
        return resp_model

    def get_details(self, data: LinePayTransactionDetail) -> LinePayResponse:
        path = "/v3/payments"
        query = ""
        if data.transaction_id is not None:
            query += "transactionId={}&".format(str(data.transaction_id))
        if data.order_id is not None:
            query += "orderId={}".format(data.order_id)
        if query.endswith("?") or query.endswith("&"):
            query = query[:-1]

        url = (
            query
            and f"{self.payment_host}{path}?{query}"
            or f"{self.payment_host}{path}"
        )

        headers = self.build_headers(path, query)
        try:
            resp = requests.get(
                url=url,
                headers=headers,
            )
            resp.raise_for_status()
            resp_json = resp.json()
            resp_model = LinePayResponse.validate(resp_json)
        except requests.HTTPError as exc:
            resp_json = {"error": exc.response.json()}
            resp_model = LinePayResponse.validate(resp_json)

        assert isinstance(resp_model, LinePayResponse)
        return resp_model



from typing import List, Literal, Optional

from pydantic import BaseModel


class B2BItem(BaseModel):
    ItemSeq: int
    ItemName: str
    ItemCount: int
    ItemWord: Optional[str]
    ItemPrice: float
    ItemAmount: float
    ItemTax: Optional[Literal["1", "2"]]
    ItemRemark: Optional[str]


class IssueB2BModel(BaseModel):
    RelateNumber: str
    InvoiceTime: Optional[str]
    CustomerIdentifier: str
    CustomerEmail: Optional[str]
    ClearanceMark: Optional[str]
    CustomerAddress: Optional[str]
    InvType: Literal["07", "08"] = "07"
    TaxType: Literal["1", "2", "3", "4"] = "1"
    TaxRate: Optional[float]
    Items: List[B2BItem]
    SalesAmount: int
    TaxAmount: int
    TotalAmount: int
    InvoiceRemark: Optional[str]


class B2CItem(BaseModel):
    ItemSeq: Optional[int]
    ItemName: str
    ItemCount: int
    ItemWord: str
    ItemPrice: int
    ItemTaxType: Optional[Literal["1", "2", "3"]]
    ItemAmount: int
    ItemRemark: Optional[str]


class IssueB2CModel(BaseModel):
    RelateNumber: str
    CustomerID: Optional[str]
    CustomerIdentifier: Optional[str]
    CustomerName: Optional[str]
    CustomerAddr: Optional[str]
    CustomerPhone: Optional[str]
    CustomerEmail: Optional[str]
    ClearanceMark: Optional[str]
    Print: Literal["0", "1"] = "0"
    Donation: Literal["0", "1"] = "0"
    LoveCode: Optional[str]
    CarrierType: Optional[Literal["1", "2", "3", ""]]
    CarrierNum: Optional[str]
    TaxType: Literal["1", "2", "3", "4", "9"] = "1"
    SpecialTaxType: Optional[Literal["1", "2", "3", "4", "5", "6", "7", "8"]]
    SalesAmount: int
    InvoiceRemark: Optional[str]
    Items: List[B2CItem]
    InvType: Literal["07", "08"]
    vat: Optional[Literal["0", "1"]]


class QueryB2BModel(BaseModel):
    InvoiceCategory: Literal[0, 1] = 0
    InvoiceNumber: str
    InvoiceDate: str
    RelateNumber: Optional[str]


class QueryB2CModel(BaseModel):
    RelateNumber: Optional[str]
    InvoiceNo: Optional[str]
    InvoiceDate: Optional[str]


class VoidB2BModel(BaseModel):
    InvoiceNumber: str
    InvoiceDate: str
    Reason: str
    Remark: Optional[str]


class VoidB2CModel(BaseModel):
    InvoiceNo: str
    InvoiceDate: str
    Reason: str



from datetime import datetime
from typing import Any, Dict, Optional, Tuple, Union
from uuid import uuid4

import boto3
from fastapi import Request, Response
from fastapi.responses import HTMLResponse, RedirectResponse

from fastel.cart import Cart
from fastel.cart.datastructures import (
    CVSLogisticSubTypes,
    InvoiceTypes,
    PaymentSubTypes,
)
from fastel.collections import get_collection, get_site_config
from fastel.config import SdkConfig
from fastel.exceptions import APIException
from fastel.logistics.ecpay.models import LogisticsTypes
from fastel.payment.common.models.callback import CallbackModel
from fastel.payment.common.models.invoice import InvoiceIssueResp, InvoiceStatus
from fastel.payment.common.models.order import Checkout, Order
from fastel.payment.ecpay import (
    ECPAY_CALLBACK_SUBTYPE,
    generate_B2B_invoice_data,
    generate_B2C_invoice_data,
)
from fastel.payment.ecpay.gateway import Ecpay, EcpayInvoice
from fastel.payment.ecpay.models.checkout import (
    EcCallback,
    EcCheckoutModel,
    EcGetNoCallback,
)
from fastel.payment.linepay.gateway import LinePay
from fastel.payment.linepay.models.checkout import (
    LinePayCallbackModel,
    LinePayCheckoutModel,
    LinePayConfirmModel,
    LinePayConfirmResp,
    LinePayRequestResp,
)
from fastel.payment.neweb.gateway import NewebInvoice, Newebpay
from fastel.payment.neweb.models.callback import (
    CallbackMsg,
    EncryptedCallback,
    GetNumMsg,
)
from fastel.payment.neweb.models.checkout import CheckoutModel
from fastel.payment.neweb.models.invoice import IssueModel, IssueResp
from fastel.payment.utils import PaymentStatus, auto_order_number, parse_order_number
from fastel.utils import TW, default_timestamp, timestamp_postfix
from fastel.utils.requests import parse_formdata


# Neweb Helpers
class NewebpayCheckout:
    def __init__(self, callback_url: str, number_url: str, redirect_url: str) -> None:
        self.gateway = Newebpay(
            SdkConfig.neweb_merchant_id,
            SdkConfig.neweb_hash_key,
            SdkConfig.neweb_hash_iv,
            stage=SdkConfig.stage,
        )
        self.notify_url = callback_url
        self.customer_url = number_url
        self.return_url = redirect_url

    def __call__(
        self, cart: Cart, return_obj: bool = False, **kwargs: Any
    ) -> Union[HTMLResponse, Tuple[HTMLResponse, Dict[str, Any]]]:
        checkout_model = self.parse_payload(cart, **kwargs)
        now = default_timestamp()
        checkout_ins = {
            "order_number": checkout_model.MerchantOrderNo,
            "created": now,
            "updated": now,
            "expired": now + 7 * 24 * 3600 * 1000,
            "attempt": 0,
            "resolved": False,
        }
        checkout_obj = cart.to_checkout(**checkout_ins)
        if return_obj:
            return (
                HTMLResponse(self.gateway.checkout_request(checkout_model)),
                checkout_obj,
            )
        return HTMLResponse(self.gateway.checkout_request(checkout_model))

    @staticmethod
    def _replace_item_name(name: str, count: int) -> str:
        result = len(name) > 30 and name[:30] + "..." or name
        if count > 1:
            result += f"等{count}項商品"
        return result

    def parse_payload(self, cart: Cart, **kwargs: Any) -> CheckoutModel:
        type_map: Dict[str, Any] = {}
        if cart.payment_subtype == PaymentSubTypes.credit:
            type_map["CREDIT"] = 1
        elif cart.payment_subtype == PaymentSubTypes.atm:
            type_map["VACC"] = 1
        elif cart.payment_subtype == PaymentSubTypes.linepay:
            type_map["LINEPAY"] = 1
        elif cart.payment_subtype == PaymentSubTypes.webatm:
            type_map["WEBATM"] = 1
        if get_site_config(key="cvs_provider", default="ecpay") == "neweb":
            if cart.logistics_type == LogisticsTypes.CVS.value:
                if cart.logistics_subtype == CVSLogisticSubTypes.FAMIC2C.value:
                    type_map["LgsType"] = "C2C"
                else:
                    type_map["LgsType"] = "B2C"
                if cart.total > 20000 or cart.total < 30:
                    raise APIException(
                        status_code=400,
                        error="validation_error",
                        detail="物流訂單商品不能超過20000元新台幣或低於30元",
                    )
                if cart.payment_subtype == PaymentSubTypes.cod.value:
                    type_map["CVSCOM"] = 2
                else:
                    type_map["CVSCOM"] = 1
        items = cart.items.items

        return CheckoutModel(
            Amt=cart.total,
            ItemDesc=self._replace_item_name(name=items[0].name, count=len(items)),
            MerchantOrderNo=auto_order_number(),
            Email=cart.buyer_email,
            NotifyURL=self.notify_url,
            CustomerURL=self.customer_url,
            ReturnURL=self.return_url,
            **type_map,
            **kwargs,
        )


class NewebpayResume:
    def __init__(self, callback_url: str, number_url: str, redirect_url: str) -> None:
        self.gateway = Newebpay(
            SdkConfig.neweb_merchant_id,
            SdkConfig.neweb_hash_key,
            SdkConfig.neweb_hash_iv,
            stage=SdkConfig.stage,
        )
        self.notify_url = callback_url
        self.customer_url = number_url
        self.return_url = redirect_url

    def __call__(self, checkout: Checkout, **kwargs: Any) -> HTMLResponse:
        checkout_model = self.parse_payload(checkout, **kwargs)
        return HTMLResponse(self.gateway.checkout_request(checkout_model))

    def parse_payload(self, checkout: Checkout, **kwargs: Any) -> CheckoutModel:
        type_map = {}
        if checkout.payment_subtype == PaymentSubTypes.credit:
            type_map["CREDIT"] = 1
        elif checkout.payment_subtype == PaymentSubTypes.atm:
            type_map["VACC"] = 1
        elif checkout.payment_subtype == PaymentSubTypes.linepay:
            type_map["LINEPAY"] = 1
        elif checkout.payment_subtype == PaymentSubTypes.cod:
            type_map["CVSCOM"] = 1

        get_collection("checkout").update_one(
            {"order_number": checkout.order_number}, {"$inc": {"attempt": 1}}
        )
        return CheckoutModel(
            Amt=checkout.total,
            ItemDesc=checkout.items[0].name,
            MerchantOrderNo=checkout.order_number + "{:04d}".format(checkout.attempt),
            Email=checkout.buyer_email,
            NotifyURL=self.notify_url,
            CustomerURL=self.customer_url,
            ReturnURL=self.return_url,
            **type_map,
            **kwargs,
        )


class NewebpayCallback:
    def __init__(self) -> None:
        self.gateway = Newebpay(
            SdkConfig.neweb_merchant_id,
            SdkConfig.neweb_hash_key,
            SdkConfig.neweb_hash_iv,
            stage=SdkConfig.stage,
        )

    def __call__(self, data: Dict[str, Any]) -> CallbackModel:
        validated = self.validate(data)
        decrypted = self.gateway.decrypt_callback(validated)
        parsed = self.parse_payload(decrypted)
        return parsed

    async def parse_request(self, request: Request) -> Dict[str, Any]:
        data = await request.form()
        return parse_formdata(data)

    def validate(self, data: Dict[str, Any]) -> EncryptedCallback:
        return EncryptedCallback.validate(data)

    def parse_payload(self, decrypted: CallbackMsg) -> CallbackModel:
        if decrypted.Status == "SUCCESS":
            payment_status = PaymentStatus.SUCCESS
        else:
            payment_status = PaymentStatus.FAILURE

        result = decrypted.Result
        if result.PaymentType == "CREDIT":
            payment_subtype = PaymentSubTypes.credit
        elif result.PaymentType == "VACC":
            payment_subtype = PaymentSubTypes.atm
        elif result.PaymentType == "CVSCOM":
            payment_subtype = PaymentSubTypes.cod
        elif result.PaymentType == "BARCODE":
            payment_subtype = PaymentSubTypes.barcode
        elif result.PaymentType == "WEBATM":
            payment_subtype = PaymentSubTypes.webatm
        else:
            payment_subtype = PaymentSubTypes.unknown

        parsered_order_number = parse_order_number(decrypted.Result.MerchantOrderNo)
        return CallbackModel(
            payment_status=payment_status,
            order_id=parsered_order_number,
            payment_subtype=payment_subtype,
            payment_transaction_detail=result.dict(),
            total=result.Amt,
        )

    def respond(self, **kwargs: Any) -> Response:
        return Response(content="1")


class NewebpayGetNumber:
    def __init__(self) -> None:
        self.gateway = Newebpay(
            SdkConfig.neweb_merchant_id,
            SdkConfig.neweb_hash_key,
            SdkConfig.neweb_hash_iv,
            stage=SdkConfig.stage,
        )

    def __call__(self, data: Dict[str, Any]) -> CallbackModel:
        validated = self.validate(data)
        decrypted = self.gateway.decrypt_getnum(validated)
        parsed = self.parse_payload(decrypted)
        return parsed

    async def parse_request(self, request: Request) -> Dict[str, Any]:
        data = await request.form()
        return parse_formdata(data)

    def validate(self, data: Dict[str, Any]) -> EncryptedCallback:
        return EncryptedCallback.validate(data)

    def parse_payload(self, decrypted: GetNumMsg) -> CallbackModel:
        result = decrypted.Result
        if decrypted.Status == "SUCCESS":
            payment_status = PaymentStatus.CODE_GENERATED
        else:
            payment_status = PaymentStatus.FAILURE

        expire_date = result.ExpireDate
        expire_time = result.ExpireTime

        if result.PaymentType == "VACC":
            payment_subtype = PaymentSubTypes.atm
        elif result.PaymentType == "BARCODE":
            payment_subtype = PaymentSubTypes.barcode
        elif result.PaymentType == "CVS":
            payment_subtype = PaymentSubTypes.cvs

        bank_code: Optional[str] = ""
        bank_account: Optional[str] = ""

        code_no: Optional[str] = ""

        barcode_1: Optional[str] = ""
        barcode_2: Optional[str] = ""
        barcode_3: Optional[str] = ""

        if payment_subtype == PaymentSubTypes.atm:
            bank_code = result.BankCode
            bank_account = result.CodeNo
        elif payment_subtype == PaymentSubTypes.cvs:
            code_no = result.CodeNo
        elif payment_subtype == PaymentSubTypes.barcode:
            barcode_1 = result.Barcode_1
            barcode_2 = result.Barcode_2
            barcode_3 = result.Barcode_3

        return CallbackModel(
            payment_status=payment_status,
            order_id=result.MerchantOrderNo,
            payment_redirect_detail=result.dict(exclude_none=True),
            total=result.Amt,
            pay_deadline=f"{expire_date} {expire_time}",
            payment_subtype=payment_subtype,
            bank_code=bank_code,
            bank_account=bank_account,
            code_no=code_no,
            barcode_1=barcode_1,
            barcode_2=barcode_2,
            barcode_3=barcode_3,
        )

    def respond(self, **kwargs: Any) -> Response:
        order_id = kwargs["order_id"]
        checkout = get_collection("checkout").find_one({"order_number": order_id})
        redirect_url = f"{SdkConfig.web_host}/profile/orders/?action=detail&is_created=true&order_number={order_id}"
        if checkout["owner"].startswith("tmp_"):
            redirect_url = f"{SdkConfig.web_host}/anonymous/order/?action=detail&is_created=true&order_number={order_id}&tmp_token={checkout['owner']}"
        return RedirectResponse(
            url=redirect_url,
            status_code=302,
        )


class NewebInvoiceIssue:
    def __init__(self) -> None:
        self.gateway = NewebInvoice(
            merchant_id=SdkConfig.neweb_invoice_merchant_id,
            hash_key=SdkConfig.neweb_invoice_hash_key,
            hash_iv=SdkConfig.neweb_invoice_hash_iv,
            stage=SdkConfig.stage,
        )

    def __call__(self, order: Dict[str, Any]) -> InvoiceIssueResp:
        validated = self.validate(order)
        parsed = self.parse_order(validated)
        gateway_resp = self.gateway.issue(parsed)
        return self.parse_gateway_resp(gateway_resp)

    def parse_gateway_resp(self, gateway_resp: IssueResp) -> InvoiceIssueResp:
        if gateway_resp.Status == "SUCCESS":
            status = InvoiceStatus.success
        else:
            status = InvoiceStatus.failure

        if gateway_resp.Result:
            invoice_number = gateway_resp.Result.InvoiceNumber
            invoice_random_no = gateway_resp.Result.RandomNum
        else:
            invoice_number = ""
            invoice_random_no = ""

        return InvoiceIssueResp(
            status=status,
            invoice_message=gateway_resp.Message,
            invoice_number=invoice_number,
            invoice_detail=gateway_resp.dict(),
            invoice_random_no=invoice_random_no,
        )

    def validate(self, order: Dict[str, Any]) -> Order:
        return Order.validate(order)

    def parse_order(self, order: Order) -> IssueModel:
        # items = order.items + order.extra_items + order.discount_items
        item_name = "|".join(list(map(lambda item: item.name, order.items)))
        item_count = "|".join(list(map(lambda item: str(item.config.qty), order.items)))
        item_unit = "|".join(list(map(lambda item: "個", order.items)))
        _amount = ""
        _price = ""
        for extra_item in order.extra_items:
            item_name += "|" + extra_item.get("name", "")
            item_count += "|1"
            item_unit += "|個"
            if order.invoice_type == InvoiceTypes.B2B:
                _amount += "|" + str(round(extra_item.get("amount", 0) / 1.05, 2))
                _price += "|" + str(round(extra_item.get("amount", 0) / 1.05, 2))
            else:
                _amount += "|" + str(extra_item.get("amount", 0))
                _price += "|" + str(extra_item.get("amount", 0))
        for discount_item in order.discount_items:
            item_name += "|" + discount_item.get("name", "")
            item_count += "|1"
            item_unit += "|個"
            if order.invoice_type == InvoiceTypes.B2B:
                _amount += "|-" + str(round(discount_item.get("sale_amount", 0), 2))
                _price += "|-" + str(round(discount_item.get("sale_amount", 0), 2))
            else:
                _amount += "|-" + str(discount_item.get("amount", 0))
                _price += "|-" + str(discount_item.get("amount", 0))
        if order.invoice_type == InvoiceTypes.B2B:
            item_amount = "|".join(
                list(map(lambda item: str(round(item.sales_amount)), order.items))
            )
            item_price = "|".join(
                list(map(lambda item: str(round(item.unit_sales)), order.items))
            )
        else:
            item_amount = "|".join(
                list(map(lambda item: str(item.amount), order.items))
            )
            item_price = "|".join(list(map(lambda item: str(item.price), order.items)))
        item_amount += _amount
        item_price += _price
        payload: Dict[str, Any] = {"BuyerName": order.buyer_name}

        if order.invoice_type == InvoiceTypes.B2B:
            payload["Category"] = "B2B"
            payload["BuyerUBN"] = order.b2b_company_no
            payload["BuyerName"] = order.b2b_company_name
            payload["PrintFlag"] = "Y"

        elif order.invoice_type == InvoiceTypes.B2C_DONATE:
            payload["Category"] = "B2C"
            payload["LoveCode"] = order.b2c_donate_code
            payload["PrintFlag"] = "N"

        elif order.invoice_type == InvoiceTypes.B2C_PHONE_CARRIER:
            payload["Category"] = "B2C"
            payload["CarrierType"] = "0"
            payload["CarrierNum"] = order.b2c_phone_carrier_code
            payload["PrintFlag"] = "N"

        elif order.invoice_type == InvoiceTypes.B2C_NPC:
            payload["Category"] = "B2C"
            payload["CarrierType"] = "1"
            payload["CarrierNum"] = order.b2c_npc_code
            payload["PrintFlag"] = "N"

        elif order.invoice_type == InvoiceTypes.B2C_PROVIDER:
            payload["Category"] = "B2C"
            payload["CarrierType"] = "2"
            payload["CarrierNum"] = order.buyer_email
            payload["PrintFlag"] = "N"

        elif order.invoice_type == InvoiceTypes.B2C:
            payload["Category"] = "B2C"
            payload["PrintFlag"] = "Y"

        return IssueModel(
            Status="1",
            MerchantOrderNo=auto_order_number(),
            Amt=order.sales,
            TaxAmt=order.tax,
            TotalAmt=order.total,
            TaxRate="5",
            TaxType="1",
            ItemName=item_name,
            ItemCount=item_count,
            ItemUnit=item_unit,
            ItemPrice=item_price,
            ItemAmt=item_amount,
            BuyerEmail=order.buyer_email,
            **payload,
        )


# Ecpay Helpers
class EcpayCheckout:
    def __init__(self, callback_url: str, number_url: str, redirect_url: str) -> None:
        self.gateway = Ecpay(
            SdkConfig.ecpay_merchant_id,
            SdkConfig.ecpay_hash_key,
            SdkConfig.ecpay_hash_iv,
            stage=SdkConfig.stage,
        )
        self.return_url = callback_url
        self.payment_info_url = number_url
        self.order_result_url = redirect_url

    def __call__(
        self, cart: Cart, return_obj: bool = False, **kwargs: Any
    ) -> Union[HTMLResponse, Tuple[HTMLResponse, Dict[str, Any]]]:
        checkout_model = self.parse_payload(cart, **kwargs)
        now = default_timestamp()
        checkout_ins = {
            "order_number": checkout_model.MerchantTradeNo,
            "created": now,
            "updated": now,
            "expired": now + 7 * 24 * 3600 * 1000,
            "attempt": 0,
            "resolved": False,
        }
        checkout_obj = cart.to_checkout(**checkout_ins)
        if return_obj:
            return (
                HTMLResponse(self.gateway.checkout_request(checkout_model)),
                checkout_obj,
            )
        return HTMLResponse(self.gateway.checkout_request(checkout_model))

    @staticmethod
    def _replace_limit_name(name: str, limit: int) -> str:
        if len(name) > limit:
            return name[: limit - 3] + "..."
        return name

    def parse_payload(self, cart: Cart, **kwargs: Any) -> EcCheckoutModel:
        now = datetime.now(TW)
        type_map = {}
        if cart.payment_subtype == PaymentSubTypes.credit:
            type_map["ChoosePayment"] = "Credit"
            type_map["OrderResultURL"] = self.order_result_url
        elif cart.payment_subtype == PaymentSubTypes.webatm:
            type_map["ChoosePayment"] = "WebATM"
            type_map["OrderResultURL"] = self.order_result_url
        elif cart.payment_subtype == PaymentSubTypes.atm:
            type_map["ChoosePayment"] = "ATM"
            type_map["PaymentInfoURL"] = self.payment_info_url
            type_map["ClientRedirectURL"] = self.order_result_url
        elif cart.payment_subtype == PaymentSubTypes.cvs:
            type_map["ChoosePayment"] = "CVS"
            type_map["PaymentInfoURL"] = self.payment_info_url
            type_map["ClientRedirectURL"] = self.order_result_url
        elif cart.payment_subtype == PaymentSubTypes.barcode:
            type_map["ChoosePayment"] = "BARCODE"
            type_map["PaymentInfoURL"] = self.payment_info_url
            type_map["ClientRedirectURL"] = self.order_result_url
        elif cart.payment_subtype == PaymentSubTypes.default:
            type_map["ChoosePayment"] = "ALL"
            type_map["PaymentInfoURL"] = self.payment_info_url
            type_map["ClientRedirectURL"] = self.order_result_url
            type_map["OrderResultURL"] = self.order_result_url
        else:
            raise ValueError("invalid Payment Subtypes")

        name = "#".join([item.name for item in cart.items.items])
        item_name = self._replace_limit_name(name=name, limit=400)

        return EcCheckoutModel(
            TotalAmount=cart.total,
            ItemName=item_name,
            MerchantTradeNo=auto_order_number(),
            MerchantTradeDate=now.strftime("%Y/%m/%d %H:%M:%S"),
            ReturnURL=self.return_url,
            **type_map,
            **kwargs,
        )


class EcpayResume:
    def __init__(self, callback_url: str, number_url: str, redirect_url: str) -> None:
        self.gateway = Ecpay(
            SdkConfig.ecpay_merchant_id,
            SdkConfig.ecpay_hash_key,
            SdkConfig.ecpay_hash_iv,
            stage=SdkConfig.stage,
        )
        self.return_url = callback_url
        self.payment_info_url = number_url
        self.order_result_url = redirect_url

    def __call__(self, checkout: Checkout, **kwargs: Any) -> HTMLResponse:
        checkout_model = self.parse_payload(checkout, **kwargs)
        return HTMLResponse(self.gateway.checkout_request(checkout_model))

    @staticmethod
    def _replace_limit_name(name: str, limit: int) -> str:
        if len(name) > limit:
            return name[: limit - 3] + "..."
        return name

    def parse_payload(self, checkout: Checkout, **kwargs: Any) -> EcCheckoutModel:
        now = datetime.now(TW)
        type_map = {}
        if checkout.payment_subtype == PaymentSubTypes.credit:
            type_map["ChoosePayment"] = "Credit"
            type_map["OrderResultURL"] = self.order_result_url
        elif checkout.payment_subtype == PaymentSubTypes.webatm:
            type_map["ChoosePayment"] = "WebATM"
            type_map["OrderResultURL"] = self.order_result_url
        elif checkout.payment_subtype == PaymentSubTypes.atm:
            type_map["ChoosePayment"] = "ATM"
            type_map["PaymentInfoURL"] = self.payment_info_url
            type_map["ClientRedirectURL"] = self.order_result_url
        elif checkout.payment_subtype == PaymentSubTypes.cvs:
            type_map["ChoosePayment"] = "CVS"
            type_map["PaymentInfoURL"] = self.payment_info_url
            type_map["ClientRedirectURL"] = self.order_result_url
        elif checkout.payment_subtype == PaymentSubTypes.barcode:
            type_map["ChoosePayment"] = "BARCODE"
            type_map["PaymentInfoURL"] = self.payment_info_url
            type_map["ClientRedirectURL"] = self.order_result_url
        elif checkout.payment_subtype == PaymentSubTypes.default:
            type_map["ChoosePayment"] = "ALL"
            type_map["PaymentInfoURL"] = self.payment_info_url
            type_map["ClientRedirectURL"] = self.order_result_url
            type_map["OrderResultURL"] = self.order_result_url
        else:
            raise ValueError("invalid Payment Subtypes")

        name = "#".join([item.name for item in checkout.items])
        item_name = self._replace_limit_name(name=name, limit=400)

        get_collection("checkout").update_one(
            {"order_number": checkout.order_number}, {"$inc": {"attempt": 1}}
        )

        return EcCheckoutModel(
            TotalAmount=checkout.total,
            ItemName=item_name,
            MerchantTradeNo=checkout.order_number + "{:04d}".format(checkout.attempt),
            MerchantTradeDate=now.strftime("%Y/%m/%d %H:%M:%S"),
            ReturnURL=self.return_url,
            **type_map,
            **kwargs,
        )


class EcpayCallback:
    def __init__(self) -> None:
        self.gateway = Ecpay(
            SdkConfig.ecpay_merchant_id,
            SdkConfig.ecpay_hash_key,
            SdkConfig.ecpay_hash_iv,
            stage=SdkConfig.stage,
        )
        self.response = "0|NOPE"

    def __call__(self, data: Dict[str, Any]) -> CallbackModel:
        validated = self.validate(data)
        self.check_mac_data(data=validated)
        parsed = self.parse_payload(validated)
        return parsed

    async def parse_request(self, request: Request) -> Dict[str, Any]:
        data = await request.form()
        return parse_formdata(data)

    def validate(self, data: Dict[str, Any]) -> EcCallback:
        return EcCallback.validate(data)

    def check_mac_data(self, data: EcCallback) -> None:
        mac_value = data.CheckMacValue
        result_data = data.dict(exclude_none=True)
        result_data.pop("CheckMacValue", None)

        result = self.gateway.cryptor.encrypt(result_data)
        if result == mac_value:
            self.response = "1|OK"

    def parse_payload(self, result: EcCallback) -> CallbackModel:
        if result.RtnCode == "1":
            payment_status = PaymentStatus.SUCCESS
        else:
            payment_status = PaymentStatus.FAILURE

        payment_type = result.PaymentType.split("_")[0]

        parsered_order_number = parse_order_number(result.MerchantTradeNo)

        return CallbackModel(
            payment_status=payment_status,
            order_id=parsered_order_number,
            total=result.TradeAmt,
            payment_subtype=ECPAY_CALLBACK_SUBTYPE[payment_type],
            payment_transaction_id=result.TradeNo,
            payment_transaction_time=result.PaymentDate,
            payment_transaction_detail=result.dict(exclude_none=True),
        )

    def respond(self, **kwargs: Any) -> Response:
        return Response(content=self.response)


class EcpayGetNumber:
    def __init__(self) -> None:
        self.gateway = Ecpay(
            SdkConfig.neweb_merchant_id,
            SdkConfig.neweb_hash_key,
            SdkConfig.neweb_hash_iv,
            stage=SdkConfig.stage,
        )
        self.response = "0|NOPE"

    def __call__(self, data: Dict[str, Any]) -> CallbackModel:
        validated = self.validate(data)
        self.check_mac_data(data=validated)
        parsed = self.parse_payload(validated)
        return parsed

    async def parse_request(self, request: Request) -> Dict[str, Any]:
        data = await request.form()
        return parse_formdata(data)

    def validate(self, data: Dict[str, Any]) -> EcGetNoCallback:
        return EcGetNoCallback.validate(data)

    def check_mac_data(self, data: EcGetNoCallback) -> None:
        mac_value = data.CheckMacValue
        result_data = data.dict(exclude_none=True)
        result_data.pop("CheckMacValue", None)

        if data.PaymentType.startswith("CVS"):
            result_data.pop("BankCode", None)
            result_data.pop("vAccount", None)
        else:
            result_data.pop("Barcode1")
            result_data.pop("Barcode2")
            result_data.pop("Barcode3")
            result_data.pop("PaymentNo", None)

        result = self.gateway.cryptor.encrypt(result_data)
        if result == mac_value:
            self.response = "1|OK"

    def parse_payload(self, result: EcGetNoCallback) -> CallbackModel:
        if result.RtnCode in ["2", "10100073"]:
            payment_status = PaymentStatus.CODE_GENERATED
        else:
            payment_status = PaymentStatus.FAILURE

        result_payment_type = result.PaymentType.split("_")[0]

        bank_code: Optional[str] = ""
        bank_account: Optional[str] = ""

        code_no: Optional[str] = ""

        barcode_1: Optional[str] = ""
        barcode_2: Optional[str] = ""
        barcode_3: Optional[str] = ""

        if result_payment_type == "ATM":
            bank_code = result.BankCode
            bank_account = result.vAccount
        elif result_payment_type == "BARCODE":
            barcode_1 = result.Barcode1
            barcode_2 = result.Barcode2
            barcode_3 = result.Barcode3
        elif result_payment_type == "CVS":
            code_no = result.PaymentNo

        return CallbackModel(
            payment_status=payment_status,
            order_id=result.MerchantTradeNo,
            total=result.TradeAmt,
            pay_deadline=result.ExpireDate,
            payment_subtype=ECPAY_CALLBACK_SUBTYPE[result_payment_type],
            payment_transaction_id=result.TradeNo,
            payment_transaction_time=result.TradeDate,
            payment_redirect_detail=result.dict(exclude_none=True),
            bank_code=bank_code,
            bank_account=bank_account,
            code_no=code_no,
            barcode_1=barcode_1,
            barcode_2=barcode_2,
            barcode_3=barcode_3,
        )

    def respond(self, **kwargs: Any) -> Response:
        return Response(content=self.response)


class EcpayInvoiceIssue:
    def __init__(self) -> None:
        self.gateway = EcpayInvoice(
            merchant_id=SdkConfig.ecpay_invoice_merchant_id,
            hash_key=SdkConfig.ecpay_invoice_hash_key,
            hash_iv=SdkConfig.ecpay_invoice_hash_iv,
            stage=SdkConfig.stage,
        )

    def __call__(self, order: Dict[str, Any]) -> InvoiceIssueResp:
        validated = self.validate(order)
        parsed = self.parse_order(validated)
        gateway_resp = self.gateway.issue(parsed)
        result = self.parse_gateway_resp(gateway_resp=gateway_resp)
        return result

    def validate(self, order: Dict[str, Any]) -> Order:
        return Order.validate(order)

    def parse_order(self, order: Order) -> Dict[str, Any]:
        if order.invoice_type == InvoiceTypes.B2B:
            payload = generate_B2B_invoice_data(order=order.dict())
        else:
            payload = generate_B2C_invoice_data(order=order.dict())
        payload["MerchantID"] = SdkConfig.ecpay_invoice_merchant_id

        return payload

    def parse_gateway_resp(self, gateway_resp: Dict[str, Any]) -> InvoiceIssueResp:
        resp_data = gateway_resp.get("Data", {})
        if resp_data["RtnCode"] == 1:
            status = InvoiceStatus.success
            invoice_number = resp_data.get("InvoiceNumber", "") or resp_data.get(
                "InvoiceNo", ""
            )
        else:
            status = InvoiceStatus.failure
            invoice_number = ""

        return InvoiceIssueResp(
            status=status,
            invoice_message=resp_data.get("RtnMsg", ""),
            invoice_number=invoice_number,
            invoice_detail=gateway_resp,
            invoice_random_no=resp_data.get("RandomNumber", ""),
        )


class LinePayCheckout:
    def __init__(self, callback_url: str, error_url: str) -> None:
        self.gateway = LinePay(
            SdkConfig.linepay_channel_id,
            SdkConfig.linepay_channel_secret,
            stage=SdkConfig.stage,
        )
        self.callback_url = callback_url
        self.error_url = error_url

    def __call__(
        self, cart: Cart, return_obj: bool = False, **kwargs: Any
    ) -> Union[Response, Tuple[Response, Dict[str, Any]]]:
        checkout_model = self.parse_payload(cart, **kwargs)
        checkout_resp = self.gateway.checkout_request(checkout_model)
        response = self.parse_gateway_resp(checkout_resp)
        now = default_timestamp()
        checkout_ins = {
            "order_number": checkout_model.orderId,
            "created": now,
            "updated": now,
            "expired": now + 7 * 24 * 3600 * 1000,
            "attempt": 0,
            "resolved": False,
        }
        checkout_obj = cart.to_checkout(**checkout_ins)
        if return_obj:
            return response, checkout_obj
        return response

    def parse_gateway_resp(self, gateway_resp: LinePayRequestResp) -> Response:
        if gateway_resp.returnCode == "0000" and gateway_resp.info:
            redirect_url = gateway_resp.info.paymentUrl.web
        else:
            redirect_url = self.error_url

        return RedirectResponse(url=redirect_url, status_code=302)

    def parse_payload(self, cart: Cart, **kwargs: Any) -> LinePayCheckoutModel:
        if cart.payment_subtype != PaymentSubTypes.linepay:
            raise ValueError("payment_subtype should be linepay")

        first_item = cart.items.items[0].to_dict()
        name = first_item["name"] + "與其他商品."
        checkout_products = [
            {
                "id": uuid4().__str__(),
                "amount": cart.total,
                "products": [
                    {
                        "id": uuid4().__str__(),
                        "name": name,
                        "imageUrl": first_item["product"]["images"][0]["expected_url"],
                        "quantity": 1,
                        "price": cart.total,
                    }
                ],
            }
        ]
        return LinePayCheckoutModel(
            amount=cart.total,
            currency="TWD",
            orderId=auto_order_number(),
            packages=checkout_products,
            #  cancelUrl: 使用者通過LINE付款頁，取消付款後跳轉到該URL
            redirectUrls={
                "confirmUrl": self.callback_url,
                "cancelUrl": SdkConfig.web_host,
            },
        )


class LinePayCallback:
    def __init__(self) -> None:
        self.gateway = LinePay(
            SdkConfig.linepay_channel_id,
            SdkConfig.linepay_channel_secret,
            stage=SdkConfig.stage,
        )

    def __call__(self, data: Dict[str, Any]) -> CallbackModel:
        callback_model = self.validate(data)
        confirm_model = self.build_confirm_model(callback_model)
        gateway_resp = self.gateway.checkout_confirm(confirm_model)
        return self.parse_gateway_resp(gateway_resp)

    def validate(self, data: Dict[str, Any]) -> LinePayCallbackModel:
        return LinePayCallbackModel.validate(data)

    def build_confirm_model(
        self,
        callback_model: LinePayCallbackModel,
    ) -> LinePayConfirmModel:
        checkout_dict = get_collection("checkout").find_one(
            {"order_number": callback_model.orderId}
        )
        return LinePayConfirmModel(
            amount=checkout_dict["total"], transactionId=callback_model.transactionId
        )

    async def parse_request(self, request: Request) -> Dict[str, Any]:
        data: Dict[str, Any] = request.query_params._dict
        return data

    def parse_gateway_resp(self, gateway_resp: LinePayConfirmResp) -> CallbackModel:
        if gateway_resp.returnCode == "0000" and gateway_resp.info:
            payment_status = PaymentStatus.SUCCESS
            total = gateway_resp.info.payInfo[0]["amount"]  # type: ignore
            payment_transaction_id = gateway_resp.info.transactionId
            order_id = gateway_resp.info.orderId
        else:
            payment_status = PaymentStatus.FAILURE
            total = 0
            payment_transaction_id = "error"
            order_id = "error"
        return CallbackModel(
            payment_subtype=PaymentSubTypes.linepay,
            payment_status=payment_status,
            payment_transaction_id=payment_transaction_id,
            payment_transaction_detail=gateway_resp.info,
            total=total,
            order_id=order_id,
        )

    def respond(self, **kwargs: Any) -> Response:
        order_id = kwargs["order_id"]
        return RedirectResponse(
            url=f"{SdkConfig.web_host}/profile/orders/?action=detail&is_created=true&order_number={order_id}",
            status_code=302,
        )


def get_resume_method(
    prefix_path: str = "payment",
) -> Union[NewebpayResume, EcpayResume]:
    default_payment_provider = get_site_config("default_payment_provider", "ecpay")
    if default_payment_provider == "neweb":
        return NewebpayResume(
            callback_url=f"{SdkConfig.api_host}/{prefix_path}/callback",
            number_url=f"{SdkConfig.api_host}/{prefix_path}/number",
            redirect_url=f"{SdkConfig.api_host}/{prefix_path}/return",
        )
    elif default_payment_provider == "ecpay":
        return EcpayResume(
            callback_url=f"{SdkConfig.api_host}/{prefix_path}/callback",
            number_url=f"{SdkConfig.api_host}/{prefix_path}/number",
            redirect_url=f"{SdkConfig.api_host}/{prefix_path}/return",
        )

    raise ValueError("unrecognize payment provider")


def get_checkout_method(
    prefix_path: str = "payment",
) -> Union[NewebpayCheckout, EcpayCheckout, LinePayCheckout]:
    default_payment_provider = get_site_config("default_payment_provider", "ecpay")
    if default_payment_provider == "neweb":
        return NewebpayCheckout(
            callback_url=f"{SdkConfig.api_host}/{prefix_path}/callback",
            number_url=f"{SdkConfig.api_host}/{prefix_path}/number",
            redirect_url=f"{SdkConfig.api_host}/{prefix_path}/return",
        )
    elif default_payment_provider == "ecpay":
        return EcpayCheckout(
            callback_url=f"{SdkConfig.api_host}/{prefix_path}/callback",
            number_url=f"{SdkConfig.api_host}/{prefix_path}/number",
            redirect_url=f"{SdkConfig.api_host}/{prefix_path}/return",
        )
    elif default_payment_provider == "linepay":
        return LinePayCheckout(
            callback_url=f"{SdkConfig.api_host}/{prefix_path}/callback",
            error_url=f"{SdkConfig.api_host}/misc/error/?title=結帳發生錯誤&url={SdkConfig.web_host}&btn_text=返回首頁",
        )

    raise ValueError("unrecognize payment provider")


def get_callback_method() -> Union[NewebpayCallback, EcpayCallback, LinePayCallback]:
    default_payment_provider = get_site_config("default_payment_provider", "ecpay")
    if default_payment_provider == "neweb":
        return NewebpayCallback()
    elif default_payment_provider == "ecpay":
        return EcpayCallback()
    elif default_payment_provider == "linepay":
        return LinePayCallback()
    raise ValueError("unrecognize payment provider")


def get_getnum_method() -> Union[NewebpayGetNumber, EcpayGetNumber]:
    default_payment_provider = get_site_config("default_payment_provider", "ecpay")
    if default_payment_provider == "neweb":
        return NewebpayGetNumber()
    elif default_payment_provider == "ecpay":
        return EcpayGetNumber()
    raise ValueError("unrecognize payment provider")


def get_invoice_issue_method() -> Union[NewebInvoiceIssue, EcpayInvoiceIssue]:
    default_payment_provider = get_site_config("default_payment_provider", "ecpay")
    if default_payment_provider == "neweb":
        return NewebInvoiceIssue()
    elif default_payment_provider == "ecpay":
        return EcpayInvoiceIssue()
    raise ValueError("unrecognize payment provider")


stepfn = boto3.client("stepfunctions", "ap-northeast-1")


async def process_callback(request: Request) -> Response:
    callback_method = get_callback_method()
    data = await callback_method.parse_request(request)
    callback_result = callback_method(data)
    stepfn_name = (
        f"callback_{callback_result.order_id}_{callback_result.payment_status}"
    )
    stepfn.start_execution(
        stateMachineArn=SdkConfig.payment_stepfn_arn,
        name=timestamp_postfix(stepfn_name),
        input=callback_result.json(),
    )

    return callback_method.respond(order_id=callback_result.order_id)


async def process_redirect(request: Request) -> Response:
    callback_method = get_callback_method()
    data = await callback_method.parse_request(request)
    callback_result = callback_method(data)
    checkout = get_collection("checkout").find_one(
        {"order_number": callback_result.order_id}
    )
    redirect_url = f"{SdkConfig.web_host}/profile/orders/?action=detail&is_created=true&order_number={callback_result.order_id}"
    if checkout["owner"].startswith("tmp_"):
        redirect_url = f"{SdkConfig.web_host}/anonymous/order/?action=detail&is_created=true&order_number={callback_result.order_id}&tmp_token={checkout['owner']}"
    return RedirectResponse(
        url=redirect_url,
        status_code=302,
    )


async def process_getnum(request: Request) -> Response:
    getnum_method = get_getnum_method()
    data = await getnum_method.parse_request(request)
    getnum_result = getnum_method(data)
    stepfn_name = f"getnum_{getnum_result.order_id}_{getnum_result.payment_status}"
    stepfn.start_execution(
        stateMachineArn=SdkConfig.payment_stepfn_arn,
        name=timestamp_postfix(stepfn_name),
        input=getnum_result.json(),
    )

    return getnum_method.respond(order_id=getnum_result.order_id)


def process_invoice_issue(order: Dict[str, Any]) -> InvoiceIssueResp:
    issue_method = get_invoice_issue_method()
    issue_resp = issue_method(order)

    return issue_resp



from typing import Any, Dict, List

from bson import ObjectId
from pydantic import BaseModel, Field

from fastel.cart.datastructures import CartConfig, ItemConfig, Product
from fastel.payment.utils import PaymentStatus, auto_order_number


class ProductItem(BaseModel):
    name: str
    amount: int
    price: int
    sales_amount: float
    unit_sales: float
    product: Product

    config: ItemConfig


class Checkout(CartConfig):
    id: ObjectId = Field(alias="_id")
    owner: str
    order_number: str = Field(default_factory=auto_order_number)
    total: int
    subtotal: int
    sales: int
    fee: int
    discount: int
    tax: int

    items: List[ProductItem]
    extra_items: List[Dict[str, Any]] = []
    discount_items: List[Dict[str, Any]] = []

    attempt: int = 0

    class Config:
        arbitrary_types_allowed = True


class Order(Checkout):
    order_id: str
    is_custom: bool = False
    offline_account: str = ""
    offline_date: str = ""
    offline_tx: str = ""
    offline_username: str = ""
    payment_status: PaymentStatus = PaymentStatus.PENDING
    payment_transaction_detail: Dict[str, Any] = {}
    payment_transaction_id: str = ""
    payment_transaction_time: str = ""
    invoice_status: str = "pending"
    invoice_detail: Dict[str, Any] = {}
    invoice_message: str = ""
    invoice_number: str = ""
    status: str = "waiting"
    logistics: List[Dict[str, Any]] = []



from typing import Any

from fastapi import Request
from fastapi.responses import JSONResponse


class APIException(Exception):
    def __init__(self, status_code: int, error: str, detail: Any):
        super().__init__(detail)
        self.status_code = status_code
        self.error = error
        self.detail = detail
        print("[APIException]", status_code, error, detail)


def api_exception_handler(request: Request, exc: APIException) -> JSONResponse:
    print("[api_exception_handler]", exc.status_code, exc.error, exc.detail)
    return JSONResponse(
        status_code=exc.status_code,
        content={"error": exc.error, "detail": exc.detail},
    )



from typing import List, Literal, Optional

from pydantic import BaseModel

PLATFORM_CODE = Literal[
    "INT0053",
    "INT0054",
    "INT0257",
    "INT0451",
    "INT0391",
    "INT0392",
    "INT0399",
    "INT0400",
    "INT0263",
    "INT0007",
    "INT0008",
    "INT0137",
    "INT0139",
    "INT0013",
    "INT0014",
    "INT0253",
    "INT0282",
    "INT0275",
    "INT0173",
    "INT0174",
    "INT0175",
    "INT0176",
    "INT0177",
    "INT0172",
    "INT0001",
    "INT0009",
    "INT0287",
    "INT0280",
    "INT0201",
    "INT0010",
    "INT0005",
    "INT0006",
    "INT0011",
    "INT0214",
    "INT0248",
]

COUNTRY = Literal[
    "CN",
    "HK",
    "TW",
    "MO",
    "AU",
    "BD",
    "IN",
    "ID",
    "JP",
    "KR",
    "MY",
    "NZ",
    "SG",
    "TH",
    "US",
    "VN",
    "AT",
    "BE",
    "BG",
]

CURRENCY = Literal[
    "USD",
    "CNY",
    "HKD",
    "EUR",
    "RUB",
    "NTD",
    "MOP",
    "SGD",
    "JPY",
    "KRW",
    "MYR",
    "VND",
    "THB",
    "AUD",
    "MNT",
]


class PaymentInfo(BaseModel):
    # 付款方式 1 寄方付， 2 收方付， 3 第三方付
    payMethod: Literal["1", "2", "3"] = "3"
    taxPayMethod: Literal["1", "2", "3"] = "3"
    payMonthCard: str
    taxPayMonthCard: Optional[str]


class BaseInfo(BaseModel):
    # 收/寄件人名字
    contact: str
    # 收/寄件國家/地區
    country: COUNTRY = "TW"
    # 郵編
    postCode: str
    # 州/省
    regionFirst: str = "台灣省"
    # 城市
    regionSecond: str = "台北市"
    # 區
    regionThird: str = "大同區"
    address: str
    email: str = ""
    cargoType: Literal[1, 2] = 1


class SenderInfo(BaseInfo):
    telNo: str = ""
    certType: Optional[str]
    certCardNo: Optional[str]


class ReceiverInfo(BaseInfo):
    phoneNo: str = ""


class ParcelInfoList(BaseModel):
    name: str
    unit: str = "個"
    amount: float
    currency: CURRENCY = "NTD"
    quantity: float
    originCountry: COUNTRY = "TW"


class SFLogisticsModel(BaseModel):
    customerCode: str
    orderOperateType: Literal[1, 5]
    customerOrderNo: str
    interProductCode: PLATFORM_CODE = "INT0053"
    # 包裹數 *
    parcelQuantity: int = 1
    # 聲明價值 * order
    declaredValue: int
    # 包裹总计声明价值币种
    declaredCurrency: str = "NTD"
    # 寄件方式 0: 服務點自寄或自行聯繫快遞員 1: 上門收件
    pickupType: Literal["0", "1"] = "1"
    # 上門區間預約時間 yyyy-MM-dd HH:mm 如果pickupType 為 1 則必填
    pickupAppointTime: Optional[str]
    # 收件時區
    pickupAppointTimeZone: str = "Asia/Taipei"
    # 運單備註 *
    remark: str = ""
    # 付款方式
    paymentInfo: PaymentInfo
    # 寄件人訊息
    senderInfo: SenderInfo
    # 收件人訊息
    receiverInfo: ReceiverInfo
    # 包裹訊息
    parcelInfoList: List[ParcelInfoList]



# utils
import time

import pytz

TW = pytz.timezone("Asia/Taipei")


def default_timestamp() -> int:
    return round(time.time() * 1000)


def timestamp_postfix(name: str) -> str:
    return name + "_" + str(int(time.time()))



from typing import Any, Dict

from requests import Response

from fastel.utils import requests


def api_token(sf_url: str, app_key: str, secret: str) -> str:
    url = sf_url + f"/openapi/api/token?appKey={app_key}&appSecret={secret}"
    resp: Response = requests.get(url)
    result = resp.json()
    if result["apiResultCode"] != 0:
        return ""
    result_data: Dict[str, Any] = result["apiResultData"]
    return result_data.get("accessToken", "")



from typing import Literal, Optional

from pydantic import BaseModel


class EZItem(BaseModel):
    name: str
    quantity: float
    price: float
    amount: float
    unit: str = "個"


class IssueModel(BaseModel):
    TransNum: Optional[str]
    MerchantOrderNo: str
    Status: Literal["0", "1", "3"] = "1"
    CreateStatusTime: Optional[str]
    Category: Literal["B2B", "B2C"] = "B2C"
    BuyerName: str
    BuyerUBN: Optional[str]
    BuyerAddress: Optional[str]
    BuyerEmail: Optional[str]
    ItemName: str
    ItemUnit: str
    ItemPrice: str
    ItemAmt: str
    ItemCount: str
    CarrierType: Optional[Literal["0", "1", "2"]]
    CarrierNum: Optional[str]
    LoveCode: Optional[str]
    PrintFlag: Literal["Y", "N"] = "Y"
    TaxType: Literal["1", "2", "3", "9"] = "1"
    TaxRate: int = 5  # 5%
    CustomsClearance: Optional[Literal["1", "2"]]
    Amt: int
    AmtSales: Optional[int]
    AmtZero: Optional[int]
    AmtFree: Optional[int]
    TaxAmt: int
    TotalAmt: int


class QueryModel(BaseModel):
    SearchType: Optional[Literal["0", "1"]]
    MerchantOrderNo: str
    TotalAmt: str
    InvoiceNumber: str
    RandomNum: str
    DisplayFlag: Optional[Literal["1"]]


class VoidModel(BaseModel):
    InvoiceNumber: str
    InvalidReason: str


class IssueResult(BaseModel):
    CheckCode: str
    MerchantID: str
    MerchantOrderNo: str
    InvoiceNumber: str
    TotalAmt: str
    InvoiceTransNo: str
    RandomNum: str
    CreateTime: str
    BarCode: Optional[str]
    QRcodeL: Optional[str]
    QRcodeR: Optional[str]


class IssueResp(BaseModel):
    Status: str
    Message: str
    Result: Optional[IssueResult]



import random
from typing import Any, Type

from fastel.config import SdkConfig

from .jwt_payload import JWTPayloadBuilder
from .kms_encode import kms_encode as encode
from .pem_coder import pem_encode, pem_random_key


class JWTEncoder:
    payload_builder_class: Type[JWTPayloadBuilder]
    stage: str = "stg"

    def __init__(self, client: Any) -> None:
        self.client = client

    def encode(self, user: Any, pem: bool = False) -> str:
        builder = self.payload_builder_class(client=self.client, data=user)
        ext = user.get("ext", {})
        payload = builder.get_payload()
        payload["ext"] = ext

        if pem:
            kid, private_key = pem_random_key(self.stage)
            return pem_encode(
                payload,
                private_key,
                headers={"kid": kid},
            )
        keys = SdkConfig.public_key
        key = random.choice(keys["keys"])["kid"]
        return encode(payload, key=key, headers={"kid": key})
